@@ -1,16 +1,18 @@
 import hashlib
+import json
 import random
 import string
 from collections import namedtuple
 
 import kubernetes
 import six
-import yaml
 
-from dagster import Array, DagsterInvariantViolationError, Field, Noneable, StringSource
+from dagster import Array, Field, Noneable, StringSource
 from dagster import __version__ as dagster_version
 from dagster import check
-from dagster.config.field_utils import Shape
+from dagster.config.field_utils import Permissive, Shape
+from dagster.config.validate import validate_config
+from dagster.core.errors import DagsterInvalidConfigError
 from dagster.serdes import whitelist_for_serdes
 from dagster.utils import frozentags, merge_dicts
 
@@ -33,29 +35,130 @@ DAGSTER_PG_PASSWORD_SECRET_KEY = 'postgresql-password'
 # Kubernetes Job object names cannot be longer than 63 characters
 MAX_K8S_NAME_LEN = 63
 
+# TODO: Deprecate this tag
 K8S_RESOURCE_REQUIREMENTS_KEY = 'dagster-k8s/resource_requirements'
-K8s_RESOURCE_REQUIREMENTS_VALID_KEYS = set(['limits', 'requests'])
+K8S_RESOURCE_REQUIREMENTS_SCHEMA = Shape({'limits': Permissive(), 'requests': Permissive()})
+
+USER_DEFINED_K8S_CONFIG_KEY = 'dagster-k8s/config'
+USER_DEFINED_K8S_CONFIG_SCHEMA = Shape(
+    {
+        'container_config': Permissive(),
+        'pod_template_spec_metadata': Permissive(),
+        'pod_spec_config': Permissive(),
+        'job_config': Permissive(),
+        'job_metadata': Permissive(),
+        'job_spec_config': Permissive(),
+    }
+)
+
+
+class UserDefinedDagsterK8sConfig(
+    namedtuple(
+        '_UserDefinedDagsterK8sConfig',
+        'container_config pod_template_spec_metadata pod_spec_config job_config job_metadata job_spec_config',
+    )
+):
+    def __new__(
+        cls,
+        container_config=None,
+        pod_template_spec_metadata=None,
+        pod_spec_config=None,
+        job_config=None,
+        job_metadata=None,
+        job_spec_config=None,
+    ):
+
+        container_config = check.opt_dict_param(container_config, 'container_config', key_type=str)
+        pod_template_spec_metadata = check.opt_dict_param(
+            pod_template_spec_metadata, 'pod_template_spec_metadata', key_type=str
+        )
+        pod_spec_config = check.opt_dict_param(pod_spec_config, 'pod_spec_config', key_type=str)
+        job_config = check.opt_dict_param(job_config, 'job_config', key_type=str)
+        job_metadata = check.opt_dict_param(job_metadata, 'job_metadata', key_type=str)
+        job_spec_config = check.opt_dict_param(job_spec_config, 'job_spec_config', key_type=str)
+
+        return super(UserDefinedDagsterK8sConfig, cls).__new__(
+            cls,
+            container_config=container_config,
+            pod_template_spec_metadata=pod_template_spec_metadata,
+            pod_spec_config=pod_spec_config,
+            job_config=job_config,
+            job_metadata=job_metadata,
+            job_spec_config=job_spec_config,
+        )
+
+    def to_dict(self):
+        return {
+            'container_config': self.container_config,
+            'pod_template_spec_metadata': self.pod_template_spec_metadata,
+            'pod_spec_config': self.pod_spec_config,
+            'job_config': self.job_config,
+            'job_metadata': self.job_metadata,
+            'job_spec_config': self.job_spec_config,
+        }
+
+    @classmethod
+    def from_dict(self, config_dict):
+        return UserDefinedDagsterK8sConfig(
+            container_config=config_dict.get('container_config'),
+            pod_template_spec_metadata=config_dict.get('pod_template_spec_metadata'),
+            pod_spec_config=config_dict.get('pod_spec_config'),
+            job_config=config_dict.get('job_config'),
+            job_metadata=config_dict.get('job_metadata'),
+            job_spec_config=config_dict.get('job_spec_config'),
+        )
 
 
 def get_k8s_resource_requirements(tags):
     check.inst_param(tags, 'tags', frozentags)
-    req_str = tags.get(K8S_RESOURCE_REQUIREMENTS_KEY)
-    if req_str is not None:
-        req_dict = yaml.safe_load(req_str)
+    check.invariant(K8S_RESOURCE_REQUIREMENTS_KEY in tags)
 
-        req_keys = set(req_dict.keys())
-        if len(req_keys.difference(K8s_RESOURCE_REQUIREMENTS_VALID_KEYS)) > 0:
-            raise DagsterInvariantViolationError(
-                'Invalid K8s resource specification. {resource} expected to only contain keys in '
-                'set {valid_keys} but found extra keys {extra_keys}'.format(
-                    resource=req_dict,
-                    valid_keys=K8s_RESOURCE_REQUIREMENTS_VALID_KEYS,
-                    extra_keys=req_keys.difference(K8s_RESOURCE_REQUIREMENTS_VALID_KEYS),
-                )
+    resource_requirements = json.loads(tags[K8S_RESOURCE_REQUIREMENTS_KEY])
+    result = validate_config(K8S_RESOURCE_REQUIREMENTS_SCHEMA, resource_requirements)
+
+    if not result.success:
+        raise DagsterInvalidConfigError(
+            'Error in tags for {}'.format(K8S_RESOURCE_REQUIREMENTS_KEY), result.errors, result,
+        )
+
+    return result.value
+
+
+def get_user_defined_k8s_config(tags):
+    check.inst_param(tags, 'tags', frozentags)
+
+    if not any(key in tags for key in [K8S_RESOURCE_REQUIREMENTS_KEY, USER_DEFINED_K8S_CONFIG_KEY]):
+        return UserDefinedDagsterK8sConfig()
+
+    user_defined_k8s_config = {}
+
+    if USER_DEFINED_K8S_CONFIG_KEY in tags:
+        user_defined_k8s_config_value = json.loads(tags[USER_DEFINED_K8S_CONFIG_KEY])
+        result = validate_config(USER_DEFINED_K8S_CONFIG_SCHEMA, user_defined_k8s_config_value)
+
+        if not result.success:
+            raise DagsterInvalidConfigError(
+                'Error in tags for {}'.format(USER_DEFINED_K8S_CONFIG_KEY), result.errors, result,
             )
-        return req_dict
 
-    return None
+        user_defined_k8s_config = result.value
+
+    container_config = user_defined_k8s_config.get('container_config', {})
+
+    # Backcompat for resource requirements key
+    if K8S_RESOURCE_REQUIREMENTS_KEY in tags:
+        resource_requirements_config = get_k8s_resource_requirements(tags)
+        container_config = merge_dicts(
+            container_config, {'resources': resource_requirements_config}
+        )
+
+    return UserDefinedDagsterK8sConfig(
+        container_config=container_config,
+        pod_template_spec_metadata=user_defined_k8s_config.get('pod_template_spec_metadata'),
+        pod_spec_config=user_defined_k8s_config.get('pod_spec_config'),
+        job_config=user_defined_k8s_config.get('job_config'),
+        job_spec_config=user_defined_k8s_config.get('job_spec_config'),
+    )
 
 
 def get_job_name_from_run_id(run_id):
@@ -107,7 +210,7 @@ class DagsterK8sJobConfig(
 
     def __new__(
         cls,
-        job_image,
+        job_image=None,
         dagster_home=None,
         image_pull_policy=None,
         image_pull_secrets=None,
@@ -119,7 +222,7 @@ class DagsterK8sJobConfig(
     ):
         return super(DagsterK8sJobConfig, cls).__new__(
             cls,
-            job_image=check.str_param(job_image, 'job_image'),
+            job_image=check.opt_str_param(job_image, 'job_image'),
             dagster_home=check.opt_str_param(
                 dagster_home, 'dagster_home', default=DAGSTER_HOME_DEFAULT
             ),
@@ -180,10 +283,12 @@ class DagsterK8sJobConfig(
         '''
         return {
             'job_image': Field(
-                StringSource,
-                is_required=True,
-                description='Docker image to use for launched task Jobs '
-                '(e.g. "mycompany.com/dagster-k8s-image:latest").',
+                Noneable(StringSource),
+                is_required=False,
+                description='Docker image to use for launched task Jobs. If the repository is not '
+                'loaded from a GRPC server, then this field is required. If the repository is '
+                'loaded from a GRPC server, then leave this field empty.'
+                '(Ex: "mycompany.com/dagster-k8s-image:latest").',
             ),
             'image_pull_policy': Field(
                 StringSource,
@@ -252,7 +357,14 @@ class DagsterK8sJobConfig(
 
 
 def construct_dagster_k8s_job(
-    job_config, command, args, job_name, resources=None, pod_name=None, component=None
+    job_config,
+    command,
+    args,
+    job_name,
+    user_defined_k8s_config=None,
+    pod_name=None,
+    component=None,
+    env_vars=None,
 ):
     '''Constructs a Kubernetes Job object for a dagster-graphql invocation.
 
@@ -267,6 +379,7 @@ def construct_dagster_k8s_job(
             in length. Defaults to "<job_name>-pod".
         component (str, optional): The name of the component, used to provide the Job label
             app.kubernetes.io/component. Defaults to None.
+        env_vars(Dict[str, str]): Additional environment variables to add to the K8s Container.
 
     Returns:
         kubernetes.client.V1Job: A Kubernetes Job object.
@@ -275,9 +388,16 @@ def construct_dagster_k8s_job(
     command = check.list_param(command, 'command', of_type=str)
     check.list_param(args, 'args', of_type=str)
     check.str_param(job_name, 'job_name')
-    check.opt_dict_param(resources, 'resources', key_type=str, value_type=dict)
+    user_defined_k8s_config = check.opt_inst_param(
+        user_defined_k8s_config,
+        'user_defined_k8s_config',
+        UserDefinedDagsterK8sConfig,
+        UserDefinedDagsterK8sConfig(),
+    )
+
     pod_name = check.opt_str_param(pod_name, 'pod_name', default=job_name + '-pod')
     check.opt_str_param(component, 'component')
+    check.opt_dict_param(env_vars, 'env_vars', key_type=str, value_type=str)
 
     check.invariant(
         len(job_name) <= MAX_K8S_NAME_LEN,
@@ -302,6 +422,11 @@ def construct_dagster_k8s_job(
     if component:
         dagster_labels['app.kubernetes.io/component'] = component
 
+    additional_k8s_env_vars = []
+    if env_vars:
+        for key, value in env_vars.items():
+            additional_k8s_env_vars.append(kubernetes.client.V1EnvVar(name=key, value=value))
+
     job_container = kubernetes.client.V1Container(
         name=job_name,
         image=job_config.job_image,
@@ -318,9 +443,9 @@ def construct_dagster_k8s_job(
                     )
                 ),
             ),
-        ],
+        ]
+        + additional_k8s_env_vars,
         env_from=job_config.env_from_sources,
-        resources=kubernetes.client.V1ResourceRequirements(**resources) if resources else None,
         volume_mounts=[
             kubernetes.client.V1VolumeMount(
                 name='dagster-instance',
@@ -330,6 +455,7 @@ def construct_dagster_k8s_job(
                 sub_path='dagster.yaml',
             )
         ],
+        **user_defined_k8s_config.container_config
     )
 
     config_map_volume = kubernetes.client.V1Volume(
@@ -338,7 +464,11 @@ def construct_dagster_k8s_job(
     )
 
     template = kubernetes.client.V1PodTemplateSpec(
-        metadata=kubernetes.client.V1ObjectMeta(name=pod_name, labels=dagster_labels),
+        metadata=kubernetes.client.V1ObjectMeta(
+            name=pod_name,
+            labels=dagster_labels,
+            **user_defined_k8s_config.pod_template_spec_metadata
+        ),
         spec=kubernetes.client.V1PodSpec(
             image_pull_secrets=[
                 kubernetes.client.V1LocalObjectReference(name=x['name'])
@@ -348,18 +478,23 @@ def construct_dagster_k8s_job(
             restart_policy='Never',
             containers=[job_container],
             volumes=[config_map_volume],
+            **user_defined_k8s_config.pod_spec_config
         ),
     )
 
     job = kubernetes.client.V1Job(
         api_version='batch/v1',
         kind='Job',
-        metadata=kubernetes.client.V1ObjectMeta(name=job_name, labels=dagster_labels),
+        metadata=kubernetes.client.V1ObjectMeta(
+            name=job_name, labels=dagster_labels, **user_defined_k8s_config.job_metadata
+        ),
         spec=kubernetes.client.V1JobSpec(
             template=template,
             backoff_limit=K8S_JOB_BACKOFF_LIMIT,
             ttl_seconds_after_finished=K8S_JOB_TTL_SECONDS_AFTER_FINISHED,
+            **user_defined_k8s_config.job_spec_config
         ),
+        **user_defined_k8s_config.job_config
     )
     return job
 
